/**
 * @file
 * VuoImage implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

extern "C" {
#include "node.h"
}
#include "VuoImageCommon.h"
#include "VuoGlContext.h"

#include <string.h>

#include <curl/curl.h>
#include <FreeImage.h>

/// @todo After we drop 10.6 support, switch back to gl3.h.
//#include <OpenGL/gl3.h>
#include <OpenGL/gl.h>


extern "C"
{
/// @todo when https://b33p.net/kosada/node/4018 is fixed, move dependencies from vuo.image.get.c.
//const char * moduleDependencies[] = {
//};
}


/**
 * Initializes the cURL and FreeImage libraries.
 */
__attribute__((constructor)) static void VuoImageCommon_init(void)
{
	curl_global_init(CURL_GLOBAL_DEFAULT);
	FreeImage_Initialise(true);
}

/**
 * A memory buffer, filled by @c VuoImageCommon_curlCallback().
 */
struct VuoImageCommon_curlBuffer
{
	char *memory;
	size_t size;
};

/**
 * Reads the contents of the URL into a memory buffer.
 * A callback function for use by @c curl_easy_setopt().
 */
static size_t VuoImageCommon_curlCallback(void *contents, size_t size, size_t nmemb, void *userp)
{
	size_t realsize = size * nmemb;
	struct VuoImageCommon_curlBuffer *mem = (struct VuoImageCommon_curlBuffer *)userp;

	mem->memory = (char *)realloc(mem->memory, mem->size + realsize + 1);
	if(mem->memory == NULL)
	{
		fprintf(stderr, "VuoImageCommon_curlCallback() Error: realloc() returned NULL (out of memory?).\n");
		return 0;
	}

	memcpy(&(mem->memory[mem->size]), contents, realsize);
	mem->size += realsize;
	mem->memory[mem->size] = 0;

	return realsize;
}

/**
 * @ingroup VuoImage
 * Retrieves the image at the specified @c imageURL, and creates a @c VuoImage from it.
 *
 * Automatically activates and deactivates a GL Context (if it makes it that far in the loading process).
 */
VuoImage VuoImage_get(const char *imageURL)
{
	if (!strlen(imageURL))
		return NULL;

	// Retrieve the image data from imageURL into a memory buffer
	struct VuoImageCommon_curlBuffer encodedImage = {NULL, 0};
	{
		CURL *curl;
		CURLcode res;

		curl = curl_easy_init();
		if (!curl)
		{
			fprintf(stderr, "VuoImage_get() Error: cURL initialization failed.\n");
			return NULL;
		}

		curl_easy_setopt(curl, CURLOPT_URL, imageURL);
		curl_easy_setopt(curl, CURLOPT_FOLLOWLOCATION, 1L);

		curl_easy_setopt(curl, CURLOPT_WRITEFUNCTION, VuoImageCommon_curlCallback);
		curl_easy_setopt(curl, CURLOPT_WRITEDATA, (void *)&encodedImage);


		res = curl_easy_perform(curl);
		if(res != CURLE_OK)
		{
			fprintf(stderr, "VuoImage_get() Error: cURL request failed: %s\n", curl_easy_strerror(res));
			return NULL;
		}

		curl_easy_cleanup(curl);
	}


	// Decode the memory buffer into a straightforward array of BGRA pixels
	FIBITMAP *dib;
	unsigned char *pixels;
	{
		FIMEMORY *hmem = FreeImage_OpenMemory((BYTE *)encodedImage.memory, encodedImage.size);

		FREE_IMAGE_FORMAT fif = FreeImage_GetFileTypeFromMemory(hmem, 0);
		if (fif == FIF_UNKNOWN)
		{
			fprintf(stderr, "VuoImage_get() Error: Couldn't determine image type.\n");
			return NULL;
		}
		if (!FreeImage_FIFSupportsReading(fif))
		{
			fprintf(stderr, "VuoImage_get() Error: This image type doesn't support reading.\n");
			return NULL;
		}

		dib = FreeImage_LoadFromMemory(fif, hmem, 0);
		FreeImage_CloseMemory(hmem);

		if (!dib)
		{
			fprintf(stderr, "VuoImage_get() Error: Failed to read image.\n");
			return NULL;
		}

		if (FreeImage_GetBPP(dib) != 32)
		{
			//fprintf(stderr, "VuoImage_get() Converting image to 32 bits per pixel.\n");
			FIBITMAP *dibConverted = FreeImage_ConvertTo32Bits(dib);
			FreeImage_Unload(dib);
			dib = dibConverted;
		}

		pixels = FreeImage_GetBits(dib);
		if (!pixels)
		{
			fprintf(stderr, "VuoImage_get() Error: Couldn't get pixels from image.\n");
			FreeImage_Unload(dib);
			return NULL;
		}
	}

	unsigned long pixelsWide = FreeImage_GetWidth(dib);
	unsigned long pixelsHigh = FreeImage_GetHeight(dib);

	// Premultiply alpha
	for (unsigned int i = 0; i < pixelsWide*pixelsHigh*4; i += 4)
	{
		unsigned int a = pixels[i+3];
		pixels[i  ] = ((unsigned int)pixels[i  ] * a) / 255;
		pixels[i+1] = ((unsigned int)pixels[i+1] * a) / 255;
		pixels[i+2] = ((unsigned int)pixels[i+2] * a) / 255;
	}

	// Upload the texture to GPU memory
	GLuint glTextureName;
	VuoGlContext_use();
	{
		glGenTextures(1, &glTextureName);
		glBindTexture(GL_TEXTURE_2D, glTextureName);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, pixelsWide, pixelsHigh, 0, GL_BGRA, GL_UNSIGNED_BYTE, (GLvoid *)pixels);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	}
	VuoGlContext_disuse();

	FreeImage_Unload(dib);
	free(encodedImage.memory);

	return VuoImage_make(glTextureName, pixelsWide, pixelsHigh);
}
