/**
 * @file
 * VuoRendererColors implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code is proprietary, and may not be redistributed.
 * For more information, see http://vuo.org/license.
 */

#include "VuoRendererColors.hh"

/**
 * Creates a new color scheme provider, optionally tinted with @c tintColor.
 * If @c isSelected is true, the colors are also tinted dark blue to indicate selection.
 * If @c isHovered is true, the colors are also slightly tinted dark blue to indicate potential for selection.
 * If @c isHighlighed is true, the colors are also tinted with light blue (more easily visible at a distance) to indicate potential for cable connection.
 */
VuoRendererColors::VuoRendererColors(VuoNode::TintColor tintColor, bool isSelected, bool isHovered, bool isHighlighted)
{
	this->tintColor = tintColor;
	this->isSelected = isSelected;
	this->isHovered = isHovered;
	this->isHighlighted = isHighlighted;
}

VuoRendererColors *VuoRendererColors::sharedColors = NULL;

/**
 * Returns a shared, untinted color scheme provider.
 * Don't free this object.
 */
VuoRendererColors *VuoRendererColors::getSharedColors(void)
{
	if (!sharedColors)
		sharedColors = new VuoRendererColors();

	return sharedColors;
}

/**
 * Returns the color for the background of the composition canvas.
 * Can be overridden by @c VuoRendererComposition::setBackgroundTransparent.
 */
QColor VuoRendererColors::canvasFill(void)
{
	return QColor::fromHslF(0, 0, 1, 1);
}

/**
 * Returns the color for the background of the main section of the node (the background of the area the port labels are drawn on).
 * Also used for the background of collapsed typecast ports.
 */
QColor VuoRendererColors::nodeFill(void)
{
	return tint(QColor::fromHslF(0, 0, 3./4., 3/4.));
}

/**
 * Returns the color for the node's outline and the background of its title bar.
 * Also used for collapsed typecast port borders.
 */
QColor VuoRendererColors::nodeFrame(void)
{
	return tint(QColor::fromHslF(0, 0, 1./2., 3./4.));
}

/**
 * Returns the color for the node's title text in the node's title bar.
 */
QColor VuoRendererColors::nodeTitle(void)
{
	return tint(QColor::fromHslF(0, 0, 1, 7./8.));
}

/**
 * Returns the color for the node's class name text in the node's title bar.
 */
QColor VuoRendererColors::nodeClass(void)
{
	return tint(QColor::fromHslF(0, 0, 1, .5));
}

/**
 * Returns the color for the background of ports in the main section of the node.
 */
QColor VuoRendererColors::portFill(void)
{
	return tint(QColor::fromHslF(0, 0, 3./4., 3./8.));
}

/**
 * Returns the color for the background of ports in the titlebar of the node.
 */
QColor VuoRendererColors::portTitlebarFill(void)
{
	return nodeFrame();
}

/**
 * Returns the color for the title text of ports in the main section of the node.
 * Also used for constant value flag text.
 */
QColor VuoRendererColors::portTitle(void)
{
	return tint(QColor::fromHslF(0, 0, 1./4., 1.));
}

/**
 * Returns the color for the highlight on the upper side of left-to-right cables (and lower side of right-to-left cables).
 */
QColor VuoRendererColors::cableUpper(void)
{
	return tint(QColor::fromHslF(0, 0, .75, .5), 2.);
}

/**
 * Returns the color for the main cable body.
 */
QColor VuoRendererColors::cableMain(void)
{
	return tint(QColor::fromHslF(0, 0, .25, .5), 2.);
}

/**
 * Returns the linear interpolation in RGB colorspace at position @c t (0..1) between colors @c v0 and @c v1.
 * Uses alpha channel from @c v0.
 */
QColor VuoRendererColors::lerpColor(QColor v0, QColor v1, float t)
{
	qreal r0,g0,b0,a;
	v0.getRgbF(&r0,&g0,&b0,&a);

	qreal r1,g1,b1;
	v1.getRgbF(&r1,&g1,&b1);

	return QColor::fromRgbF(
				fmin(r0 + (r1 - r0)*t, 1.),
				fmin(g0 + (g1 - g0)*t, 1.),
				fmin(b0 + (b1 - b0)*t, 1.),
				a
				);
}

/**
 * Tints the specified color by @c tintColor.
 * @c amount is used to tint cables more strongly, since their small size makes the tint less apparent.
 */
QColor VuoRendererColors::tint(QColor color, qreal amount)
{
	qreal h,s,l,a;
	color.getHslF(&h,&s,&l,&a);

	qreal hs;
	if (tintColor == VuoNode::TintYellow)
		hs = 45./360.;
	else if (tintColor == VuoNode::TintOrange)
		hs = 15./360.;
	else if (tintColor == VuoNode::TintMagenta)
		hs = 315./360.;
	else if (tintColor == VuoNode::TintViolet)
		hs = 255./360.;
	else if (tintColor == VuoNode::TintCyan)
		hs = 175./360.;
	else if (tintColor == VuoNode::TintGreen)
		hs = 90./360.;

	if (tintColor != VuoNode::TintNone)
	{
		// Factor in the source hue only if source saturation is nonzero
		if (s > .1)
			h = h + (hs - h) * 3./4.;
		else
			h = hs;

		s = s + (1. - s) * 3./8. * amount;
	}

	color = QColor::fromHslF(h,s,l,a);

	if (isHighlighted)
	{
		QColor highlight = QColor::fromHslF(235./360., 1., .7, 1.); // Light Blue
		color = lerpColor(color, highlight, 7./8. * amount);
	}

	if (isSelected || isHovered)
	{
		QColor selection = QColor::fromHslF(235./360., 1., .3, 1.); // Dark Blue
		color = lerpColor(color, selection, (isSelected * 4./8. + isHovered * 3./8.) * amount);
	}

	return color;
}
